//
// Project : ThreadPool
// File    : thread_pool.cc
// Purpose : class for managing a pool of threads
//
// arch-tag: d4739323-4e22-42d9-9fcb-f98f5890d77b
//

#include <assert.h>
#include <pthread.h>

#include "TThreadPool.hh"

//
// set to one to enable sequential execution
//
#define THR_SEQUENTIAL  0

//
// global thread-pool
//

TThreadPool * thread_pool = NULL;

///////////////////////////////////////////////
//
// init global thread pool
//
///////////////////////////////////////////////

int
init_thread_pool ( uint p )
{
    if ( thread_pool != NULL )
        delete thread_pool;
    
    if ((thread_pool = new TThreadPool( p )) == NULL)
    {
        ERROR << "(init_thread_pool) no memory available" << std::endl;
        return -1;
    }// if

    return 0;
}

///////////////////////////////////////////////
//
// thread handled by threadpool
//
///////////////////////////////////////////////

//
// running method
//
void
TThreadPool::TPoolThr::run ()
{
    _del_mutex.lock();

#define PRINT( m )  
    
    while ( ! _end )
    {
        //
        // wait for work
        //
        
        while ((_job == NULL) && ! _end )
        {
            PRINT( std::cout << "thread " << thread_no() << " waiting" << std::endl );
            _work_mutex.lock();
            _work_mutex.unlock();
        }// while

        //
        // look if we really have a job to do
        // and handle it
        //
        
        if ( _job != NULL )
        {
            PRINT( std::cout << "thread " << thread_no() << " executing" << std::endl );
            // execute job
            _job->run( _data_ptr );
            
            // detach thread from job
            _job->set_pool_thr( NULL );

            if ( _del_job )
                delete _job;
            
            set_job( NULL, NULL );
            PRINT( std::cout << "thread " << thread_no() << " end execute" << std::endl );

            _work_mutex.lock();
            _sync_mutex.unlock();
        }// if

        // append thread to idle-list
        _pool->append_idle( this );
    }// while

    PRINT( std::cout << "thread " << thread_no() << " finishing" << std::endl );
    
    _del_mutex.unlock();
}

///////////////////////////////////////////////
//
// ThreadPool 
//
///////////////////////////////////////////////

//
// constructor and destructor
//

TThreadPool::TThreadPool ( uint max_p )
{
    //
    // create max_p threads for pool
    //

    _max_parallel = max_p;

    _threads = new TPoolThr*[ max_p ];

    for ( uint i = 0; i < max_p; i++ )
    {
        _threads[i] = new TPoolThr( i, this );
        _idle_threads.append( _threads[i] );
        _threads[i]->start( true, true );
    }// for

    // tell the scheduling system, how many threads to expect
    pthread_setconcurrency( max_p + pthread_getconcurrency() );
}

TThreadPool::~TThreadPool ()
{
    // wait till all threads have finished
    sync_all();
    
    // finish all thread
    for ( uint i = 0; i < _max_parallel; i++ )
    {
        _threads[i]->sync_mutex().lock();

        _threads[i]->set_end( true );
        _threads[i]->set_job( NULL, NULL );
        
        if ( _threads[i]->work_mutex().is_locked() )
            _threads[i]->work_mutex().unlock();

        // might have been unlocked by pool-thread
        if ( _threads[i]->sync_mutex().is_locked() )
            _threads[i]->sync_mutex().unlock();
    }// for
    
    // cancel and delete all threads (not really safe !)
    for ( uint i = 0; i < _max_parallel; i++ )
    {
        _threads[i]->del_mutex().lock();
        delete _threads[i];
    }// for

    delete[] _threads;
}

///////////////////////////////////////////////
//
// run, stop and synch with job
//

void
TThreadPool::run ( TThreadPool::TJob * job, void * ptr, bool del )
{
    assert( job != NULL );

#if THR_SEQUENTIAL == 1
    //
    // run in calling thread
    //
    
    job->run( ptr );

    if ( del )
        delete job;
    
#else
    //
    // run in concurrent thread
    //

    PRINT( std::cout << "1" << std::endl );
    TPoolThr * t = get_idle();
    
    //
    // and start the job
    //
    
    PRINT( std::cout << "2" << std::endl );
    t->sync_mutex().lock();

    // attach job to thread
    PRINT( std::cout << "3" << std::endl );
    t->set_job( job, ptr, del );

    // attach thread to job
    PRINT( std::cout << "4" << std::endl );
    job->set_pool_thr( t );

    PRINT( std::cout << "5" << std::endl );
    if ( t->work_mutex().is_locked() )
        t->work_mutex().unlock();
    PRINT( std::cout << "6" << std::endl );
#endif
}

void
TThreadPool::sync ( TJob * job )
{
    if ( job == NULL )
        return;
    
    TPoolThr  * t = job->pool_thr();

    // check if job is already released
    if ( t == NULL )
        return;
    
    //
    // look if thread is working and wait for signal
    //

    t->sync_mutex().lock();
    
    t->set_job( NULL, NULL );

    // might have been unlocked by pool-thread
    if ( t->sync_mutex().is_locked() )
        t->sync_mutex().unlock();

    // detach job and thread
    job->set_pool_thr( NULL );
}

//
// sync with all running threads
//
void
TThreadPool::sync_all ()
{
    for ( uint i = 0; i < _max_parallel; i++ )
    {
        if ( _threads[i]->sync_mutex().is_locked() )
        {
            _threads[i]->sync_mutex().lock();

            // might have been unlocked by pool-thread
            if ( _threads[i]->sync_mutex().is_locked() )
                _threads[i]->sync_mutex().unlock();
        }// if
    }// for
}

///////////////////////////////////////////////
//
// manage pool threads
//

//
// return idle thread form pool
//
TThreadPool::TPoolThr *
TThreadPool::get_idle ()
{
    while ( true )
    {
        //
        // wait for an idle thread
        //

        while ( _idle_threads.size() == 0 )
        {
            PRINT( std::cout << "waiting for idle thread" << std::endl );
            _idle_cond.wait( _idle_mutex );
        }// while

        //
        // get first idle thread
        //
        
        _list_mutex.lock();
        
        if ( _idle_threads.size() > 0 )
        {
            TPoolThr * t = _idle_threads.behead();

            _list_mutex.unlock();
            
            return t;
        }// if

        _list_mutex.unlock();
    }// while
}

void
TThreadPool::append_idle ( TThreadPool::TPoolThr * t )
{
    _list_mutex.lock();

    //
    // check if given thread is already in list
    // and only append if not so
    //
    
    TSLL< TPoolThr * >::TIterator  iter = _idle_threads.first();

    while ( ! iter.eol() )
    {
        if ( iter() == t )
        {
            _list_mutex.unlock();
            return;
        }// if

        ++iter;
    }// while
    
    _idle_threads.append( t );
    _list_mutex.unlock();
    
    _idle_cond.broadcast();
}
